import fs from 'fs';
import request from 'request';

export interface IEsouiPublishBase {
  id: number;
  version?: string;
  description?: string;
  changelog?: string;
  compatibility?: string;
}

export interface IEsouiUpdateParams extends IEsouiPublishBase {
  updatefile?: string;
  testDeploy?: boolean;
}

export interface IEsouiUpdatePackage extends IEsouiPublishBase {
  updatefile?: fs.ReadStream;
}

export interface IEsouiUpdateResult {
  responseStatusCode: number;
  responseBody: any;
  params: IEsouiUpdateParams;
}

export class EsouiPublish {
  private readonly token: string;

  private readonly params: IEsouiUpdateParams;

  constructor(token: string, params: IEsouiUpdateParams) {
    this.token = token;
    this.params = params;
  }

  /**
   * Convenience method that generates the update package and performs the update to esoui.
   **/
  async createUpdatePackageAndUpdate(): Promise<IEsouiUpdateResult> {
    return this.createEsouiUpdatePackage()
      .then(updatePackage => this.updateEsoui(updatePackage));
  }

  /**
   * Generates an 'update package' that contains the info needed to update to esoui.com
   */
  async createEsouiUpdatePackage(): Promise<IEsouiUpdatePackage> {
    const updatePackage: IEsouiUpdatePackage = {
      id: this.params.id
    }

    if (this.params.version) {
      updatePackage.version = this.params.version;
    }

    if (this.params.description) {
      updatePackage.description = await EsouiPublish.readFile(this.params.description);
    }

    if (this.params.changelog) {
      updatePackage.changelog = await EsouiPublish.readFile(this.params.changelog);
    }

    if (this.params.compatibility) {
      updatePackage.compatibility = this.params.compatibility;
    }

    if (this.params.updatefile) {
      updatePackage.updatefile = fs.createReadStream(this.params.updatefile);
    }

    return updatePackage;
  }

  /**
   * Performs update to esoui.
   * 
   * @param updatePackage 
   * 
   */
  async updateEsoui(updatePackage: IEsouiUpdatePackage): Promise<IEsouiUpdateResult> {
    const endpoint = `https://api.esoui.com/addons/update${this.params.testDeploy ? 'test' : ''}`;
    if (this.params.testDeploy) {
      console.warn('****TEST DEPLOYMENT****');
    }

    console.log(`Publishing to ${endpoint}`);
    return new Promise<IEsouiUpdateResult>((resolve, reject) => {
      const handleResponse = (error: any, httpResponse: request.Response, body: any): void => {
        if (error) {
          return reject(error);
        }

        if (httpResponse.statusCode !== 202) {
          if (httpResponse.statusCode === 403) {
            return reject('Unauthorized. Ensure correct token is set in `ESOUI_TOKEN` environment variable or provided via command line (--token=<you\'re token>)');
          } else {
            return reject(body);
          }
        }

        resolve({ responseStatusCode: httpResponse.statusCode, responseBody: JSON.parse(body), params: this.params });
      }

      request.post({
        headers: {
          'x-api-token': this.token
        },
        url: endpoint,
        formData: updatePackage
      }, handleResponse)
    });
  }

  static async publish(token: string, id: number, version?: string, description?: string, changelog?: string, compatibility?: string, updateFile?: string, test?: boolean): Promise<IEsouiUpdateResult> {
    const params: IEsouiUpdateParams = {
      id: id,
      version,
      description,
      changelog,
      compatibility,
      updatefile: updateFile,
      testDeploy: test
    }

    return new EsouiPublish(token, params).createUpdatePackageAndUpdate();
  }

  static async publishFromJson(jsonFile: string): Promise<IEsouiUpdateResult> {
    return new Promise(async (resolve, reject) => {
      let json: any;
      try {
        json = JSON.parse(await EsouiPublish.readFile(jsonFile));
      } catch (error) {
        return reject(error);
      }

      if (json && json.id) {
        const token = json.token || process.env.ESOUI_TOKEN;
        if (!token) {
          return reject('`token` not set in the JSON file or the ESOUI_TOKEN environment variable');
        }

        const params: IEsouiUpdateParams = { id: json.id };
        json.version ? params.version = json.version : undefined;
        json.description ? params.description = json.description : undefined;
        json.changelog ? params.changelog = json.changelog : undefined;
        json.compatibility ? params.compatibility = json.compatibility : undefined;
        json.updateFile ? params.updatefile = json.updateFile : undefined;
        json.test ? params.testDeploy = json.test : undefined;
        resolve(new EsouiPublish(token, params).createUpdatePackageAndUpdate());
      } else {
        return reject('Missing `id`');
      }
    });
  }

  private static async readFile(file: string): Promise<string> {
    return new Promise((resolve, reject) => fs.readFile(file, 'utf8', (error, data) => {
      if (error) {
        reject(error);
      } else {
        resolve(data);
      }
    }));
  }
}
